/*
 * mISDN HW interface
 *
 * Copyright (C) 2009, Stefan Knoblich <s.knoblich@axsentis.de>, axsentis GmbH
 *
 * NOTE: This is intended as a Layer 1 interface only,
 *       signaling is completely handled by other modules!
 */

#include <errno.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <poll.h>

#include <linux/mISDNif.h>

#include <openzap.h>

#ifndef AF_ISDN
#define AF_ISDN         34
#define PF_ISDN         AF_ISDN
#endif

/***********************************************************************************
 * mISDN interface functions
 ***********************************************************************************/

/**
 * \brief	Send a control command to a mISDN device
 */
static zap_status_t misdn_send_command(zap_socket_t sock, int command)
{
	char cmdbuf[MAX_DATA_MEM];
	struct mISDNhead *hh = (struct mISDNhead *)cmdbuf;
	int retval, len;

	memset(cmdbuf, 0, sizeof(cmdbuf));
	hh->prim = command;
	hh->id   = MISDN_ID_ANY;
	len      = MISDN_HEADER_LEN;

	retval = sendto(sock, cmdbuf, len, 0, NULL, 0);
	if (retval <= 0) {
		return ZAP_FAIL;
	}

	zap_log(ZAP_LOG_DEBUG, "mISDN sent command %d of %d bytes size (retval: %d)\n", command, len, retval);

	return ZAP_SUCCESS;
}

#define MISDN_DEFAULT_CMD_TIMEOUT	250	/*!< 250ms */

static zap_status_t misdn_recv_command(zap_socket_t sock)
{
	char cmdbuf[MAX_DATA_MEM] = { 0 };
	struct mISDNhead *hh = (struct mISDNhead *)cmdbuf;
	int retval, len = sizeof(cmdbuf);

	retval = recvfrom(sock, cmdbuf, len, 0, NULL, 0);
	if (retval <= 0) {
		return ZAP_FAIL;
	}

	zap_log(ZAP_LOG_DEBUG, "mISDN received response %d (retval: %d)\n", hh->prim, retval);

	return hh->prim;
}

/**
 * En-/Disable Echo canceller
 */
static zap_status_t misdn_set_ec(zap_channel_t *zchan, int enable)
{
	return ZAP_FAIL;
}

/***********************************************************************************
 * OpenZAP interface functions
 ***********************************************************************************/

/**
 * \brief	Open channel
 * \param	zchan	OpenZAP channel to open
 */
static ZIO_OPEN_FUNCTION(misdn_open)
{
	zap_status_t ret = 0;

	/*
	 * Send activation request
	 */
	ret = misdn_send_command(zchan->sockfd, PH_ACTIVATE_REQ);
	if (ret != ZAP_SUCCESS) {
		zap_log(ZAP_LOG_ERROR, "Failed to activate channel %d:%d: %s\n", zchan->span_id, zchan->chan_id, strerror(errno));
		return ZAP_FAIL;
	}

retry:
	ret = misdn_recv_command(zchan->sockfd);
	if (ret == PH_ACTIVATE_REQ) {
		goto retry;
	}
	if (ret != PH_ACTIVATE_IND) {
		zap_log(ZAP_LOG_ERROR, "Failed to activate channel %d:%d: %s (%d)\n", zchan->span_id, zchan->chan_id, strerror(errno), ret);
		return ZAP_FAIL;
	}

	zap_log(ZAP_LOG_NOTICE, "mISDN channel %d:%d activated\n", zchan->span_id, zchan->chan_id);

	return ZAP_SUCCESS;
}

/**
 * \brief	Close channel
 * \param	zchan	OpenZAP channel to close
 */
static ZIO_CLOSE_FUNCTION(misdn_close)
{
	zap_status_t ret = 0;

	/*
	 * Send deactivation request
	 */
	ret = misdn_send_command(zchan->sockfd, PH_DEACTIVATE_REQ);
	if (ret != ZAP_SUCCESS) {
		zap_log(ZAP_LOG_ERROR, "Failed to deactivate channel %d:%d\n", zchan->span_id, zchan->chan_id);
		return ZAP_FAIL;
	}

	ret = misdn_recv_command(zchan->sockfd);
	if (ret != PH_DEACTIVATE_IND) {
		zap_log(ZAP_LOG_ERROR, "Failed to deactivate channel %d:%d\n", zchan->span_id, zchan->chan_id);
		return ZAP_FAIL;
	}

	zap_log(ZAP_LOG_NOTICE, "mISDN channel %d:%d deactivated\n", zchan->span_id, zchan->chan_id);

	return ZAP_SUCCESS;
}

/**
 * \brief	Execute command
 * \param	zchan	OpenZAP channel
 * \param	command	Command to execute
 * \param	obj	Additional command data
 */
static ZIO_COMMAND_FUNCTION(misdn_command)
{
	switch (command) {
	case ZAP_COMMAND_ENABLE_ECHOCANCEL:
	case ZAP_COMMAND_DISABLE_ECHOCANCEL:
		if (zchan->type != ZAP_CHAN_TYPE_B)
			return ZAP_FAIL;

		return misdn_set_ec(zchan, (command == ZAP_COMMAND_ENABLE_ECHOCANCEL) ? 1 : 0);
		break;

	default:
		zap_log(ZAP_LOG_ERROR, "Unknown command %d\n", command);
	}
	return ZAP_FAIL;
}

/**
 * \brief	Wait for new data
 * \param	zchan	OpenZAP channel to wait on
 * \param	flags	Wait flags
 * \param	to	Timeout
 */
static ZIO_WAIT_FUNCTION(misdn_wait)
{
	struct pollfd pfd[1];
	int retval;
	int evt = ZAP_NO_FLAGS;

	if (*flags & ZAP_READ)
		evt |= POLLIN;
	if (*flags & ZAP_WRITE)
		evt |= POLLOUT;
	if (*flags & ZAP_EVENTS)
		evt |= POLLPRI;

	*flags = ZAP_NO_FLAGS;
	pfd[0].fd = zchan->sockfd;
	pfd[0].events = evt;

	retval = poll(pfd, 1, to);
	if (!retval) {
		return ZAP_TIMEOUT;
	}
	else if (retval < 0 || (pfd[0].revents & POLLERR)) {
		snprintf(zchan->last_error, sizeof(zchan->last_error), "Poll failed");
		return ZAP_FAIL;
	}
	else {
		evt = pfd[0].revents;
		if (evt & POLLIN)
			*flags |= ZAP_READ;
		if (evt & POLLOUT)
			*flags |= ZAP_WRITE;
		if (evt & POLLPRI)
			*flags |= ZAP_EVENTS;
	}
	return ZAP_SUCCESS;
}

/**
 * \brief	Read data
 * \param	zchan	OpenZAP channel
 * \param	data	Buffer for data
 * \param	datalen	Number of bytes to read (contains bytes read after return)
 */
static ZIO_READ_FUNCTION(misdn_read)
{
	int bytes = *datalen;
	int retries = 10;
	int retval;

	while (retries-- > 0) {
		if ((retval = read(zchan->sockfd, data, bytes)) > 0) {
			bytes = retval;
			break;
		}
	}
	if (bytes > 0) {
		*datalen = bytes;
		return ZAP_SUCCESS;
	}
	return ZAP_FAIL;
}

/**
 * \brief	Write data
 * \param	zchan	OpenZAP channel
 * \param	data	Buffer for data
 * \param	datalen	Number of bytes to write (contains bytes written after return)
 */
static ZIO_WRITE_FUNCTION(misdn_write)
{
	int bytes = *datalen;
	int retries = 10;
	int retval;

	while (retries-- > 0) {
		if ((retval = write(zchan->sockfd, data, bytes)) > 0) {
			bytes = retval;
			break;
		}
	}
	if (bytes > 0) {
		*datalen = bytes;
		return ZAP_SUCCESS;
	}
	return ZAP_FAIL;
}

static zap_status_t misdn_open_range(zap_span_t *span, zap_chan_type_t type, int card_id, int start, int end)
{
	int num_configured = 0;
	int x;

	zap_log(ZAP_LOG_DEBUG, "mISDN configuring card:range %d:%d->%d\n", card_id, start, end);

	/* channel loop */
	for (x = start; x <= end; x++) {
		zap_channel_t *zchan = NULL;
		zap_socket_t sockfd = -1;
		struct sockaddr_mISDN addr;

		zap_log(ZAP_LOG_DEBUG, "mISDN configuring card:channel => %d:%d\n", card_id, x);

		switch (type) {
		case ZAP_CHAN_TYPE_DQ921:
		case ZAP_CHAN_TYPE_DQ931:
			sockfd = socket(AF_ISDN, SOCK_DGRAM, (span->trunk_type == ZAP_TRUNK_E1) ? ISDN_P_TE_E1: ISDN_P_TE_S0);	/* TODO: protocol */
			break;

		case ZAP_CHAN_TYPE_B:
			sockfd = socket(AF_ISDN, SOCK_DGRAM, ISDN_P_B_RAW);
			break;
		default:
			zap_log(ZAP_LOG_ERROR, "Invalid channel type %d\n", type);
		}

		if (sockfd < 0) {
			zap_log(ZAP_LOG_ERROR, "Failed to open socket: %s\n", strerror(errno));
			return ZAP_FAIL;
		}

		/*
		 * Bind socket to card:channel
		 */
		memset(&addr, 0, sizeof(struct sockaddr_mISDN));
		addr.family  = AF_ISDN;
		addr.dev     = card_id - 1;
		addr.channel = x - 1;

		if (bind(sockfd, (struct sockaddr *)&addr, sizeof(struct sockaddr_mISDN)) < 0) {
			zap_log(ZAP_LOG_ERROR, "Failed to bind mISDN socket [%d:%d]: %s\n", 0, 0, strerror(errno));
			return ZAP_FAIL;
		}

		/*
		 * Other setup, if required
		 */


		/*
		 * Add channel to span
		 */
		if (zap_span_add_channel(span, sockfd, type, &zchan) != ZAP_SUCCESS) {
			zap_log(ZAP_LOG_ERROR, "Failed to add mISDN zchan to span\n");
			return ZAP_FAIL;
		}

		zchan->rate = 8000;
		zchan->physical_span_id = card_id;
		zchan->physical_chan_id = x;

		if (type == ZAP_CHAN_TYPE_B) {
			zchan->native_codec = zchan->effective_codec = ZAP_CODEC_ALAW;
		} else {
			zchan->native_codec = zchan->effective_codec = ZAP_CODEC_NONE;
		}

		num_configured++;
	}

	return num_configured;
}

/**
 * \brief	Configure/open span ozmod_misdn settings
 */
static ZIO_CONFIGURE_SPAN_FUNCTION(misdn_configure_span)
{
	int card_id = 0, range_start = 0, range_end = 0;
	int res = 0;
	char *chan_str, *ptr;
	char *data = strdup(str);


	if (!(chan_str = strchr(data, ':'))) {
		zap_log(ZAP_LOG_ERROR, "Invalid configuration string: %s\n", str);
		return ZAP_FAIL;
	} else {

		/* terminate channel range */
		*chan_str++ = '\0';

		card_id = atoi(str);

		if (!(ptr = strchr(chan_str, '-'))) {
			/* single channel */
			range_start = atoi(chan_str);
			range_end   = range_start;
		}
		else {
			*ptr++ = '\0';

			/* channel range */
			range_start = atoi(chan_str);
			range_end   = atoi(ptr);
		}
	}

	if (!card_id || !range_start || !range_end) {
		zap_log(ZAP_LOG_ERROR, "Invalid configuration string\n");
		return ZAP_FAIL;
	}

	res = misdn_open_range(span, type, card_id, range_start, range_end);
	if (res <= 0) {
		zap_log(ZAP_LOG_ERROR, "mISDN failed to configure channel(s)\n");
		return ZAP_FAIL;
	}

	if (data) {
		free(data);
	}

	return res;
}

/**
 * \brief	Configure global ozmod_misdn settings
 */
static ZIO_CONFIGURE_FUNCTION(misdn_configure)
{
	return ZAP_SUCCESS;
}

/**
 * \brief	Retrieve alarm event information (if any)
 * \param	zchan	OpenZAP channel
 */
static ZIO_GET_ALARMS_FUNCTION(misdn_get_alarms)
{
	return ZAP_FAIL;
}

/**
 * \brief	Poll for new events
 * \param	span	OpenZAP span
 * \param	ms	Timeout (in ms)
 */
static ZIO_SPAN_POLL_EVENT_FUNCTION(misdn_poll_event)
{
	return ZAP_FAIL;
}

/**
 * \brief	Retrieve event
 * \param	span	OpenZAP span
 * \param	event	OpenZAP event
 */
static ZIO_SPAN_NEXT_EVENT_FUNCTION(misdn_next_event)
{
	return ZAP_FAIL;
}

/**
 * \brief	Destroy ozmod_misdn channel
 * \param	zchan	OpenZAP channel
 */
static ZIO_CHANNEL_DESTROY_FUNCTION(misdn_channel_destroy)
{
	if (zchan->sockfd >= 0) {
		close(zchan->sockfd);
		zchan->sockfd = -1;
	}
	return ZAP_SUCCESS;
}

/**
 * \brief	Destroy ozmod_misdn span
 * \param	span	OpenZAP span
 */
static ZIO_SPAN_DESTROY_FUNCTION(misdn_span_destroy)
{
	return ZAP_FAIL;
}

/**
 * \brief	ozmod_misdn interface
 */
static const zap_io_interface_t misdn_interface = {
	.name       = "misdn",
	.open       = misdn_open,
	.close      = misdn_close,
	.wait       = misdn_wait,
	.read       = misdn_read,
	.write      = misdn_write,
	.poll_event = misdn_poll_event,
	.next_event = misdn_next_event,

	.command         = misdn_command,
	.get_alarms      = misdn_get_alarms,
	.configure       = misdn_configure,
	.configure_span  = misdn_configure_span,
	.channel_destroy = misdn_channel_destroy,
	.span_destroy    = misdn_span_destroy,
};

enum {
	MISDN_CAPS_NONE = 0,

	/* D-Channel */
	MISDN_CAPS_PRI = (1 << 0),
	MISDN_CAPS_BRI = (1 << 1),
	MISDN_CAPS_UP0 = (1 << 2),
	MISDN_CAPS_NT  = (1 << 3),
	MISDN_CAPS_TE  = (1 << 4),

	/* B-Channel */
	MISDN_CAPS_RAW  = (1 << 10),
	MISDN_CAPS_HDLC = (1 << 11),
};

#define MISDN_IS_BRI(x)	(x & MISDN_CAPS_BRI)
#define MISDN_IS_PRI(x)	(x & MISDN_CAPS_PRI)

#define MISDN_IS_TE(x)	(x & MISDN_CAPS_TE)
#define MISDN_IS_NT(x)	(x & MISDN_CAPS_NT)

#define MISDN_IS_RAW(x)  (x & MISDN_CAPS_RAW)
#define MISDN_IS_HDLC(x) (x & MISDN_CAPS_HDLC)

/**
 * \brief	ozmod_misdn module init function
 */
static ZIO_IO_LOAD_FUNCTION(misdn_load)
{
	struct mISDNversion ver;
	struct mISDN_devinfo devinfo;
	int i, devcnt, usecnt;
	zap_socket_t sock = -1;

	zap_log(ZAP_LOG_DEBUG, "Trying to open mISDN raw socket\n");

	sock = socket(AF_ISDN, SOCK_RAW, ISDN_P_BASE);
	if (sock < 0) {
		zap_log(ZAP_LOG_CRIT, "Unable to create mISDN socket (are you sure this kernel has mISDN support?)\n");
		return ZAP_FAIL;
	}

	if (ioctl(sock, IMGETVERSION, &ver) < 0) {
		zap_log(ZAP_LOG_CRIT, "Unable to retrieve mISDN version\n");
		goto error;
	}

	zap_log(ZAP_LOG_INFO, "mISDN Interface version %hhd.%hhd.%hd\n", ver.major, ver.minor, ver.release);

	devcnt = 0;
	if (ioctl(sock, IMGETCOUNT, &devcnt) < 0) {
		zap_log(ZAP_LOG_CRIT, "Unable to retrieve number of mISDN devices\n");
		goto error;

	}

	if (!devcnt) {
		zap_log(ZAP_LOG_CRIT, "No mISDN devices found\n");
		goto error;
	}
	usecnt = devcnt;

	zap_log(ZAP_LOG_INFO, "Found %d mISDN devices:\n", devcnt);

	/* Output most important device information */
	for (i = 0; i < devcnt; i++) {
		int caps = MISDN_CAPS_NONE;

		devinfo.id = i;
		if (ioctl(sock, IMGETDEVINFO, &devinfo) < 0) {
			zap_log(ZAP_LOG_ERROR, "Failed to retrieve information for device %d\n", i);
			continue;
		}

		/* print */
		zap_log(ZAP_LOG_INFO, "Name: %s, B-Channels: %d\n",
					zap_strlen_zero_buf(devinfo.name) ? "Unknown" : devinfo.name,
					devinfo.nrbchan);

		/* D-Channels capabilities */
		if (devinfo.Dprotocols & (1 << ISDN_P_TE_E1)) {
			caps |= MISDN_CAPS_TE | MISDN_CAPS_PRI;
		}
		if (devinfo.Dprotocols & (1 << ISDN_P_NT_E1)) {
			caps |= MISDN_CAPS_NT | MISDN_CAPS_PRI;
		}
		if (devinfo.Dprotocols & (1 << ISDN_P_TE_S0)) {
			caps |= MISDN_CAPS_TE | MISDN_CAPS_BRI;
		}
		if (devinfo.Dprotocols & (1 << ISDN_P_NT_S0)) {
			caps |= MISDN_CAPS_NT | MISDN_CAPS_BRI;
		}
#ifdef ISDN_P_TE_UP0
		if (devinfo.Dprotocols & (1 << ISDN_P_TE_UP0)) {
			caps |= MISDN_CAPS_TE | MISDN_CAPS_UP0 | MISDN_CAPS_BRI;
		}
#endif
#ifdef ISDN_P_NT_UP0
		if (devinfo.Dprotocols & (1 << ISDN_P_NT_UP0)) {
			caps |= MISDN_CAPS_NT | MISDN_CAPS_UP0 | MISDN_CAPS_BRI;
		}
#endif

		/* B-Channel capabilities */
		if (devinfo.Bprotocols & (1 << (ISDN_P_B_RAW & ISDN_P_B_MASK))) {
			caps |= MISDN_CAPS_RAW;
		}
		if (devinfo.Bprotocols & (1 << (ISDN_P_B_HDLC & ISDN_P_B_MASK))) {
			caps |= MISDN_CAPS_HDLC;
		}

		zap_log(ZAP_LOG_INFO, "    Type: %s, Modes: %s %s\n",
					MISDN_IS_PRI(caps) ? "PRI" : "BRI",
					MISDN_IS_NT(caps)  ? "NT" : "",
					MISDN_IS_TE(caps)  ? "TE" : "");

		zap_log(ZAP_LOG_INFO, "    B-Channel modes: %s %s\n",
					MISDN_IS_RAW(caps) ? "RAW" : "",
					MISDN_IS_HDLC(caps) ? "HDLC" : "");

		if (!(MISDN_IS_NT(caps) || MISDN_IS_TE(caps)) && !MISDN_IS_RAW(caps)) {
			zap_log(ZAP_LOG_ERROR, "   This device is unusable!\n");
			usecnt--;
		}
	}
	if (!usecnt) {
		zap_log(ZAP_LOG_CRIT, "No useable devices found!\n");
		goto error;
	}

	zap_log(ZAP_LOG_INFO, "Found %d useable mISDN devices\n", usecnt);

	/* assign interface struct */
	*zio = (zap_io_interface_t *)&misdn_interface;

	close(sock);
	return ZAP_SUCCESS;

error:
	if (sock >= 0) {
		close(sock);
	}
	return ZAP_FAIL;
}

/**
 * \brief	ozmod_misdn module shutdown
 */
static ZIO_IO_UNLOAD_FUNCTION(misdn_unload)
{
	return ZAP_SUCCESS;
}

/**
 * \brief	ozmod_misdn module
 */
zap_module_t zap_module = {
	.name      = "misdn",
	.io_load   = misdn_load,
	.io_unload = misdn_unload
};
